/* vim: cin ts=8 sw=8 tw=80 syn=c
 *
 * Copyright (c) 2013 Garry T. Williams
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <linux/limits.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/inotify.h>
#include <sys/ioctl.h>
#include <poll.h>
#include <errno.h>
#include <unistd.h>
/* #include "backlight.h" */

#define ENABLE "/sys/class/backlight/intel_backlight/brightness"
#define DEVICE "/sys/class/backlight/acpi_video0/brightness"

/*
 * The XPS-13 out of the box works flawlessly with Ubuntu.
 * Unfortunately, I want to run Fedora instead, but the backlight keys
 * and power management backlight adjustments don't work with Fedora 18
 * (3.9.2-200.fc18.x86_64).  The OSD display shows the change to
 * brightness, but the backlight remains at maximum brightness.
 *
 * If you write a "0" to /sys/class/backlight/intel_backlight/brightness,
 * the backlight immediately adjusts to the new value.  Subsequent
 * changes seem to work for a while without touching this file, but
 * eventually, the machine goes back to ignoring any change to the
 * backlight value.
 *
 * The work-around here is to run a daemon at boot time that simply
 * watches for changes in the brightness value and, when it sees a
 * change, the daemon writes a "0" to intel_backlight/brightness file.
 * This is effective in that all the changes (in KDE desktop at least)
 * to backlight values actually affect the display.
 */

int main(int argc, char *argv[]) {
	int     in_fd;
	int     watch_descriptor;
	FILE   *enable;
	char    buf[sizeof(struct inotify_event) + NAME_MAX + 1];

	/*
	 * This daemon is expected to be started so early in the boot process
	 * that the device won't be available.  We wait for the device here.
	 */
	do {
		enable = fopen(ENABLE, "r+");
		if (enable == NULL) {
			if (errno != ENOENT) {
				perror("opening device");
				exit(1);
			}

			sleep(1);
		}
	} while (enable == NULL);

	(void)fclose(enable);

	if ((in_fd = inotify_init()) < 0) {
		perror("inotify_init");
		exit(1);
	}

	/* set up to be notified on file modification */
	watch_descriptor = inotify_add_watch(in_fd, DEVICE, IN_MODIFY);
	if (watch_descriptor < 0)
	{
		perror("inotify_add_watch");
		exit(1);
	}

	/*
	 * Monitor changes to actual brightness and then write to
	 * intel_backlight seems to actually enable the backlight change that
	 * was done.
	 */
	while (1) {
		ssize_t n;

		enable = fopen(ENABLE, "r+");
		if (enable == NULL) {
			perror("open");
			exit(1);
		}

		if (fwrite("0\n", 2, 1, enable) < 0) {
			perror("fwrite");
			exit(1);
		}

		(void)fclose(enable);

	       	n = read(in_fd, buf, sizeof(buf));
		if (n < 0) {
			perror("read");
			exit(1);
		}
	}
}
